/* This file is part of the ili2fme project.
 * For more information, please see <http://www.interlis.ch/>.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
package ch.interlis.ili2fme;

import java.util.ArrayList;
import java.util.HashSet;
import java.util.HashMap;
import java.util.Iterator;

import COM.safe.fme.pluginbuilder.IFMEReader;
import COM.safe.fmeobjects.IFMEFeature;
import COM.safe.fmeobjects.IFMEFeatureVector;
import COM.safe.fme.pluginbuilder.IFMEMappingFile;
import COM.safe.fmeobjects.IFMESession;
import COM.safe.fmeobjects.FMEException;
import COM.safe.fmeobjects.IFMELogFile;

import ch.ehi.basics.logging.EhiLogger;

import ch.interlis.ili2c.metamodel.*;

import ch.interlis.iom.swig.iom_javaConstants;
import ch.interlis.iom.*;

/** INTERLIS implementation of an FME-Reader.
 * @author ce
 */
public class Ili2Reader implements IFMEReader {
	private IFMEMappingFile mappingFile=null;
	private IFMESession session=null;
	private String readerKeyword=null;
	private ch.interlis.ili2c.metamodel.TransferDescription iliTd=null;
	private ArrayList iliModelv=null;
	private HashMap tag2class=null;
	private String xtfFile=null; // null if ili-file given
	private IomIterator iomBasketi=null;
	private IomBasket iomBasket=null;
	private IomIterator iomObji=null;
	private IomObject iomObj=null;
	private IomFile iomFile=null;
	private boolean surfaceAsPolyline=false;
	public Ili2Reader(IFMESession session1,IFMEMappingFile mappingFile1,String keyword,IFMELogFile log){
		mappingFile=mappingFile1;
		readerKeyword=keyword;
		session=session1;
		setupLogging(log);
	}
	public void open(ArrayList args) throws Exception {
		if(args.size()==0){
			throw new IllegalArgumentException("args.size()==0");
		}
		String models=null;
		String modeldir=null;
		for(int i=0;i<args.size();i++){
			String arg=(String)args.get(i);
			EhiLogger.traceState("arg["+Integer.toString(i)+"] "+arg);
			if(arg.equals(Ili2fme.MODELS)){
				i++;
				models=(String)args.get(i);
			}else if(arg.equals(Ili2fme.MODEL_DIR)){
				i++;
				modeldir=(String)args.get(i);
			}else if(arg.startsWith("+")){
				i++; // skip this and next(!) argument
			}else{
				// skip this argument
			}
		}
		Iterator elei=mappingFile.elements();
		while(elei.hasNext()){
			Object eleo=elei.next();
			EhiLogger.traceState("element "+eleo.getClass()+","+eleo);
			if(eleo instanceof ArrayList){
				ArrayList ele=(ArrayList)eleo;
				String val=(String)ele.get(0);
				if(val.equals(readerKeyword+"_"+Ili2fme.MODELS)){
					models=(String)ele.get(1);	
				}else if(val.equals(readerKeyword+"_"+Ili2fme.MODEL_DIR)){
					modeldir=(String)ele.get(1);	
				}
			}
		}
		
		if(models==null){
			//models=mappingFile.fetchString(readerKeyword+"_"+Ili2fme.MODELS);
			models="XTF";
		}
		EhiLogger.traceState("models <"+models+">");
		if(modeldir==null){
			//modeldir=mappingFile.fetchString(readerKeyword+"_"+Ili2fme.MODEL_DIR);
			modeldir=new java.io.File(session.fmeHome(),"ilimodels").getAbsolutePath();
		}
		EhiLogger.traceState("modeldir <"+modeldir+">");
		
		xtfFile=(String)args.get(0);
		if(xtfFile.length()>=2 && xtfFile.charAt(0)=='/' && xtfFile.charAt(1)=='/'){
			StringBuffer x=new StringBuffer(xtfFile);
			x.setCharAt(0,'\\');
			x.setCharAt(1,'\\');
			xtfFile=x.toString();
		}
		EhiLogger.logState("xtfFile <"+xtfFile+">");
		if(models.equals("XTF")){
			String xtfExt=ch.ehi.basics.view.GenericFileFilter.getFileExtension(new java.io.File(xtfFile)).toLowerCase();
			if(xtfExt.equals("ili")){
				ArrayList iliFilev=new ArrayList();
				iliFilev.add(xtfFile);
				xtfFile=null;
				// compile models
				iliTd=Ili2fme.compileIliFiles(iliFilev,modeldir,null);
				if(iliTd==null){
					// compiler failed
					return;
				}
			}else{
				// get model names out of transfer file
				iliModelv=new ArrayList();
				setupIom();
				iomFile=Iom.openFile(xtfFile,0,null);

				IomIterator basketi=iomFile.iteratorbasket();
				// get first basket
				IomBasket basket=basketi.nextbasket();
				// no baskets?
				if(basket==null){
					// no models!
					iliTd=null;
					return;
				}
				String model[]=basket.getbaskettag().split("\\.");
				EhiLogger.logState("model from xtf <"+model[0]+">");
				iliModelv.add(model[0]);
				// compile models
				iliTd=Ili2fme.compileIliModels(iliModelv,modeldir,null);
				if(iliTd==null){
					// compiler failed
					return;
				}
			}
		}else{
			// parse string
			iliModelv=new ArrayList(java.util.Arrays.asList(models.split(";")));
			// compile models
			iliTd=Ili2fme.compileIliModels(iliModelv,modeldir,null);
			if(iliTd==null){
				// compiler failed
				return;
			}
		}
		classv=getTransferViewables(iliTd);
		classi=null;
		tag2class=ch.interlis.ili2c.generator.XSDGenerator.getTagMap(iliTd);
		// ready to scan schema or read data
	}

	// Terminate the reader mid-stream.  Any special actions to shut
	// down a reader not finished reading data should be taken in this
	// method.  For most readers, nothing will be done.
	public void abort() throws Exception {
		cleanup();
	}

	// Close the reader after it has exhausted its data source.
	public void close() throws Exception {
		cleanup();
	}
	
	// Return the unique ID assigned to this reader.
	// All Java reader return 0.
	public int id() {
		return 0;
	}

	public IFMEFeature read(IFMEFeature ret) 
		throws Exception 
	{
		// ili-file mode?
		if(xtfFile==null){
			// no features, just schema features!
			return null;
		}
		try{
			// if first call?
			if(iomBasketi==null){
				setupIom();
				iomFile=Iom.openFile(xtfFile,0,null);
				iomBasketi=iomFile.iteratorbasket();
				// get first basket
				iomBasket=iomBasketi.nextbasket();
				// no baskets?
				if(iomBasket==null){
					// end of file reached
					return null;
				}
				iomObji=iomBasket.iteratorobject();
			}
			// no more objects in current basket and basket mapped
			if(iomBasket==null){
				// get next basket
				iomBasket=iomBasketi.nextbasket();
				// no more baskets?
				if(iomBasket==null){
					// end of file reached
					if(listener.hasSeenErrors()){
						throw new Exception("INTERLIS 2 reader failed");
					}
					return null;
				}
				iomObji=iomBasket.iteratorobject();
			}
			iomObj=iomObji.nextobject();
			// no more objects (in currentBasket)?
			if(iomObj==null){
				// map basket (to a feature! to get metainfo about basket)
				mapBasket(ret,iomBasket);
				iomBasket=null; // basket mapped
				return ret;
			}
			// ASSERT: iomObj is current object
			// translate object
			mapFeature(ret,iomObj,null);
		
			return ret;
		}catch(Exception ex){
			EhiLogger.logError(ex);
			throw ex;
		}
	}
	private void mapBasket(IFMEFeature ret,IomBasket iomBasket)
	{
		ret.setFeatureType(Ili2fme.XTF_BASKETS);
		ret.setStringAttribute(Ili2fme.XTF_TOPIC,iomBasket.getbaskettag());
		ret.setStringAttribute(Ili2fme.XTF_ID,iomBasket.getbasketoid());
	}
	private void mapFeature(IFMEFeature ret,IomObject iomObj,String prefix)
	{
		boolean isStruct=prefix!=null;
		if(!isStruct)prefix="";
		String tag=iomObj.getobjecttag();
		Viewable aclass=(Viewable)tag2class.get(tag);
		if(aclass==null){
			EhiLogger.logError("line "+iomObj.getobjectline()+": unknonw class <"+tag+">; ignored");
			return;
		}
		Viewable rootClass=getRoot(aclass);
		if(rootClass==null){
		 rootClass=aclass;
		}
		//EhiLogger.debug("aclass <"+aclass+">, root <"+rootClass+">");
		if(!isStruct){
			ret.setFeatureType(rootClass.getScopedName(null));
			ret.setStringAttribute(Ili2fme.XTF_ID,iomObj.getobjectoid());
			ret.setStringAttribute(Ili2fme.XTF_BASKET,iomBasket.getbasketoid());
		}
		AttributeDef geomattr=isStruct ? null : ((ViewableWrapper)classv.get(rootClass)).getGeomAttr();
		if(geomattr!=null){
			ret.setStringAttribute(Ili2fme.XTF_GEOMATTR,geomattr.getName());
		}
		 ret.setStringAttribute(prefix+Ili2fme.XTF_CLASS,tag);
		Iterator iter = aclass.getAttributesAndRoles2();
		boolean geometryDone=false;
		while (iter.hasNext()) {
			Viewable.TransferElement prop = (Viewable.TransferElement)iter.next();
			if (prop.obj instanceof AttributeDef) {
				AttributeDef attr = (AttributeDef) prop.obj;
				Type type = attr.getDomainResolvingAliases();
				 String attrName=attr.getName();
				if (type instanceof CompositionType){
				 int valuec=iomObj.getattrvaluecount(attrName);
				 for(int valuei=0;valuei<valuec;valuei++){
					IomObject value=iomObj.getattrobj(attrName,valuei);
					if(value!=null){
						mapFeature(ret,value,attr.getName()+"{"+Integer.toString(valuei)+"}.");
					}
				 }
				 // enqueue struct values
				 //int structc=iomObj.getattrvaluecount(attrName);
				 //for(int structi=0;structi<structc;structi++){
				 //	IomObject struct=iomObj.getattrobj(attrName,structi);
			 	 //	enqueStructValue(sqlId,sqlType,attrName,struct,structi);
				 //}
				}else if (type instanceof PolylineType){
					 if(attr!=geomattr){
						 continue;
					 }
					 geometryDone=true;
					 //ret.setStringAttribute("fme_geometry{0}", "xtf_polyline");
					 IomObject value=iomObj.getattrobj(attrName,0);
					 if(value!=null){
						 ret.setStringAttribute(Ili2fme.XTF_GEOMTYPE,"xtf_polyline");
						 setPolyline(ret,value,false);
					 }					
				 }else if(type instanceof SurfaceType){
					if(attr!=geomattr){
						 continue;
					 }
					 geometryDone=true;
					 //ret.setStringAttribute("fme_geometry{0}", "xtf_surface");
					 IomObject value=iomObj.getattrobj(attrName,0);
					 if(value!=null){
						 if(surfaceAsPolyline){
							 ret.setStringAttribute(Ili2fme.XTF_GEOMTYPE,"xtf_polyline");
						 }else{
							 ret.setStringAttribute(Ili2fme.XTF_GEOMTYPE,"xtf_surface");
						 }
						 setSurface(ret,value);
					 }
				 }else if(type instanceof AreaType){
					if(attr!=geomattr){
						 continue;
					 }
					 geometryDone=true;
					 //ret.setStringAttribute("fme_geometry{0}", "xtf_area");
					 IomObject value=iomObj.getattrobj(attrName,0);
					 if(value!=null){
						 if(surfaceAsPolyline){
							 ret.setStringAttribute(Ili2fme.XTF_GEOMTYPE,"xtf_polyline");
						 }else{
							 ret.setStringAttribute(Ili2fme.XTF_GEOMTYPE,"xtf_area");
						 }
						 setSurface(ret,value);
					 }
				 }else if(type instanceof CoordType){
					if(attr!=geomattr){
						 continue;
					 }
					 geometryDone=true;
					 //ret.setStringAttribute("fme_geometry{0}", "xtf_coord");
					 IomObject value=iomObj.getattrobj(attrName,0);
					 if(value!=null){
						 ret.setStringAttribute(Ili2fme.XTF_GEOMTYPE,"xtf_coord");
						 ret.setGeometryType(IFMEFeature.FME_GEOM_POINT);
						 addCoord(ret,value);
					 }
				}else{
				 String value=iomObj.getattrvalue(attrName);
				 if(value!=null){
					 ret.setStringAttribute(prefix+attrName, value);
				 }
				}
			}
			if(prop.obj instanceof RoleDef){
				RoleDef role = (RoleDef) prop.obj;
				String roleName=role.getName();
				// a role of an embedded association?
				if(prop.embedded){
					AssociationDef roleOwner = (AssociationDef) role.getContainer();
					if(roleOwner.getDerivedFrom()==null){
						// not just a link?
						 IomObject structvalue=iomObj.getattrobj(roleName,0);
						 if(structvalue!=null){
							if (roleOwner.getAttributes().hasNext()
								|| roleOwner.getLightweightAssociations().iterator().hasNext()) {
								 // TODO add association attributes
							}
							 String refoid=structvalue.getobjectrefoid();
							 long orderPos=structvalue.getobjectreforderpos();
							 if(orderPos!=0){
								// refoid,orderPos
								ret.setStringAttribute(prefix+roleName, refoid);
								ret.setStringAttribute(prefix+roleName+"."+Ili2fme.ORDERPOS, Long.toString(orderPos));
							 }else{
								// refoid
								ret.setStringAttribute(prefix+roleName, refoid);
							 }
						 }
					}
				}else{
				 IomObject structvalue=iomObj.getattrobj(roleName,0);
				 String refoid=structvalue.getobjectrefoid();
				 long orderPos=structvalue.getobjectreforderpos();
				 if(orderPos!=0){
					// refoid,orderPos
					ret.setStringAttribute(prefix+roleName, refoid);
					ret.setStringAttribute(prefix+roleName+"."+Ili2fme.ORDERPOS, Long.toString(orderPos));
				 }else{
					// refoid
					ret.setStringAttribute(prefix+roleName, refoid);
				 }
				}
			}
		}
	}
	private Iterator classi=null;
	private HashMap classv=null; // map<Viewable, ViewableWrapper>
	public IFMEFeature readSchema(IFMEFeature ret) throws Exception {
		try{
			// ili2c failed?
			if(iliTd==null){
				return null;
			}
			// first call?
			if(classi==null){
				// create XTF_BASKETS class
				ret.setFeatureType(Ili2fme.XTF_BASKETS);
				ret.setStringAttribute("fme_geometry{0}", "xtf_none");
				ret.setStringAttribute(Ili2fme.XTF_TOPIC,Ili2fme.ILIQNAME_TYPE);
				ret.setStringAttribute(Ili2fme.XTF_ID,Ili2fme.ID_TYPE);
				// init class iterator
				classi=classv.keySet().iterator();
				return ret;
			}
			// more classes?
			while(classi.hasNext()){
				Viewable v=(Viewable)classi.next();
				if((v instanceof Table) && !((Table)v).isIdentifiable()){
					continue;
				}
				ArrayList geomAttrAccu=new ArrayList();
				mapFeatureType(geomAttrAccu,ret,v,null);						
				AttributeDef geomattr=((ViewableWrapper)classv.get(v)).getGeomAttr();
				if(geomattr==null){
					ret.setStringAttribute("fme_geometry{0}", "xtf_none");
				}else{
					Type type=geomattr.getDomainResolvingAliases();
					if (type instanceof PolylineType){
						ret.setStringAttribute("fme_geometry{0}", "xtf_polyline");
					}else if(type instanceof SurfaceType){
						if(surfaceAsPolyline){
							ret.setStringAttribute("fme_geometry{0}", "xtf_polyline");
						}else{
							ret.setStringAttribute("fme_geometry{0}", "xtf_surface");
						}
					}else if(type instanceof AreaType){
						if(surfaceAsPolyline){
							ret.setStringAttribute("fme_geometry{0}", "xtf_polyline");
						}else{
							ret.setStringAttribute("fme_geometry{0}", "xtf_area");
						}
					}else if(type instanceof CoordType){
						ret.setStringAttribute("fme_geometry{0}", "xtf_coord");
					}else{
						throw new IllegalStateException("!(type instanceof geomType)");
					}
					for(int attri=1;attri<geomAttrAccu.size();attri++){
						geomattr=(AttributeDef)geomAttrAccu.get(attri);
						EhiLogger.logAdaption(geomattr.getScopedName(null)+": Attribute ignored; FME supports only one geometry per feature");							
					}
				}
				return ret;
			}
			return null;
		}catch(Exception ex){
			EhiLogger.logError(ex);
			throw ex;
		}
	}
	private void mapFeatureType(ArrayList geomAttrAccu,IFMEFeature ret,Viewable v,String attrNamePrefix)
	{
		boolean isStruct=attrNamePrefix!=null;
		if(!isStruct)attrNamePrefix="";
		Viewable rootClass=getRoot(v);
		if(rootClass==null){
		 rootClass=v;
		}
		//EhiLogger.debug("aclass <"+v+">, root <"+rootClass+">");
		ArrayList attrv=((ViewableWrapper)classv.get(rootClass)).getAttrv();
		if(!isStruct){
			ret.setFeatureType(v.getScopedName(null));
			ret.setStringAttribute(Ili2fme.XTF_ID,Ili2fme.ID_TYPE);
			ret.setStringAttribute(Ili2fme.XTF_BASKET,Ili2fme.ID_TYPE);
			ret.setStringAttribute(Ili2fme.XTF_GEOMATTR,Ili2fme.ILINAME_TYPE);
		}
		ret.setStringAttribute(attrNamePrefix+Ili2fme.XTF_CLASS,Ili2fme.ILIQNAME_TYPE);
		AttributeDef geomattr=null;
		for(int i=0;i<attrv.size();i++){
			Object attro=attrv.get(i);
			if(attro instanceof AttributeDef){
				AttributeDef attr=(AttributeDef)attro;
				Type type=attr.getDomainResolvingAliases();
				if (type instanceof PolylineType 
					|| type instanceof SurfaceOrAreaType
					|| type instanceof CoordType){
					if(!isStruct){
						geomAttrAccu.add(attr);		
						if(geomattr==null){
							geomattr=attr;					
						}
					}else{
						EhiLogger.logAdaption(attr.getScopedName(null)+": Attribute ignored; FME supports only one geometry per feature");
					}
				}else if(type instanceof CompositionType){
					mapFeatureType(geomAttrAccu,ret,((CompositionType)type).getComponentType(),attr.getName()+"{}.");						
				}else{
					// TODO set attribute type
					ret.setStringAttribute(attrNamePrefix+attr.getName(),"xtf_char(20)");
				}
			}else if(attro instanceof RoleDef){
				RoleDef role=(RoleDef)attro;
				ret.setStringAttribute(attrNamePrefix+role.getName(),Ili2fme.ID_TYPE);
			}
		}
	}
	public boolean spatialEnabled() throws Exception {
		return false;
	}

	public void setConstraints(IFMEFeature arg0) throws Exception {
	}
	
	//
	// helpers
	//
	private boolean suppresViewable(Viewable v){
		if(v.isAbstract()){
			return true;
		}
		if(v instanceof AssociationDef){
			AssociationDef assoc=(AssociationDef)v;
			// embedded and no attributes/embedded links?
			if(assoc.isLightweight() && 
				!assoc.getAttributes().hasNext()
				&& !assoc.getLightweightAssociations().iterator().hasNext()
				) {
				return true;
			}
			if(assoc.getDerivedFrom()!=null){
				return true;
			}
		}
		
		// do not suppress STRUCTUREs here, but in readSchema()
		
		Topic topic;
		if ((v instanceof View) && ((topic=(Topic)v.getContainer (Topic.class)) != null)
			&& !topic.isViewTopic()){
				return true;
			}
		return false;
	}
	private Viewable getRoot(Viewable v)
	{
		Viewable root=(Viewable)v.getRootExtending();
		// is root a class and defined in model INTERLIS?
		if((root instanceof Table) && ((Table)root).isIdentifiable() && root.getContainerOrSame(Model.class)==iliTd.INTERLIS){
			if((v instanceof Table) && ((Table)v).isIdentifiable() && v.getContainerOrSame(Model.class)==iliTd.INTERLIS){
				// skip it
				return v;
			}
			// use base as root that is defined outside model INTERLIS
			root=v;
			Viewable nextbase=(Viewable)v.getExtending();
			while(nextbase.getContainerOrSame(Model.class)!=iliTd.INTERLIS){
				root=nextbase;
				nextbase=(Viewable)root.getExtending();
			}
		}
		return root;
	}
	/** @return map<Viewable v, ViewableWrapper wrapper>
	 */
	private HashMap getTransferViewables(ch.interlis.ili2c.metamodel.TransferDescription td)
	{
		HashSet leaveclassv=new HashSet();
		// for all models
		Iterator modeli = td.iterator ();
		while (modeli.hasNext ())
		{
		  Object mObj = modeli.next ();
		  if (mObj instanceof Model){
			Model model=(Model)mObj;
			//EhiLogger.debug("model <"+model+">");
			Iterator topici=model.iterator();
			while(topici.hasNext()){
				Object tObj=topici.next();
				if (tObj instanceof Topic){
				  Topic topic=(Topic)tObj;
				  if(topic.isAbstract()){
				  	continue;
				  }
				  //EhiLogger.debug("topic <"+topic+">");
				  Iterator iter = topic.getViewables().iterator();
				  while (iter.hasNext())
				  {
					Object obj = iter.next();
					if(obj instanceof Viewable){
						Viewable v=(Viewable)obj;
						if(suppresViewable(v)){
							continue;
						}
						//EhiLogger.debug("leaveclass <"+v+">");
					   	leaveclassv.add(v);
					}
				  }
				}else if(tObj instanceof Viewable){
					  Viewable v=(Viewable)tObj;
					  if(suppresViewable(v)){
						  continue;
					  }
					  //EhiLogger.debug("leaveclass <"+v+">");
					  leaveclassv.add(v);
				}
			}
		  }
		}
		// find base classes and add possible extensions
		Iterator vi=leaveclassv.iterator();
		HashMap basev=new HashMap(); // map<Viewable root,HashSet<Viewable extensions>> 
		while(vi.hasNext()){
			Viewable v=(Viewable)vi.next();
			//EhiLogger.debug("leaveclass <"+v+">");
			Viewable root=getRoot(v);
			if((v instanceof Table) && ((Table)v).isIdentifiable() && v.getContainerOrSame(Model.class)==td.INTERLIS){
				// skip it
				continue;
			}
			//EhiLogger.debug("  root <"+root+">");
			if(root==null){
				if(!basev.containsKey(v)){
					basev.put(v,new HashSet());
				}
			}else{
				HashSet extv;
				if(basev.containsKey(root)){
					extv=(HashSet)basev.get(root);
				}else{
					extv=new HashSet();
					basev.put(root,extv);
				}
				while(v!=root){
					extv.add(v);
					v=(Viewable)v.getExtending();
				}
			}
		}
		// build list of attributes
		vi=basev.keySet().iterator();
		HashMap ret=new HashMap();
		while(vi.hasNext()){
			Viewable v=(Viewable)vi.next();
			//EhiLogger.debug("baseclass <"+v+">");
			ArrayList attrv=new ArrayList();
			mergeAttrs(attrv,v,true);
			Iterator exti=((HashSet)basev.get(v)).iterator();
			while(exti.hasNext()){
				Viewable ext=(Viewable)exti.next();
				//EhiLogger.debug("  ext <"+ext+">");
				mergeAttrs(attrv,ext,false);
			}
			ViewableWrapper wrapper=new ViewableWrapper();
			wrapper.setAttrv(attrv);
			wrapper.setViewable(v);
			boolean isStruct=(v instanceof Table) && !((Table)v).isIdentifiable();
			for(int i=0;i<attrv.size();i++){
				Object attro=attrv.get(i);
				if(attro instanceof AttributeDef){
					AttributeDef attr=(AttributeDef)attro;
					Type type=attr.getDomainResolvingAliases();
					if (type instanceof PolylineType 
						|| type instanceof SurfaceOrAreaType
						|| type instanceof CoordType){
						if(!isStruct){
							wrapper.setGeomAttr(attr);
							break;
						}
					}
				}
			}
			ret.put(v,wrapper);
		}
		// addSecondGeomAttrs(ret,v);
		return ret;
	}
	private void mergeAttrs(ArrayList attrv,Viewable v,boolean isRoot){
		Iterator iter = null;
		if(isRoot){
			iter=v.getAttributesAndRoles2();
		}else{
			iter=v.getDefinedAttributesAndRoles2();
		}
		while (iter.hasNext()) {
			Viewable.TransferElement obj = (Viewable.TransferElement)iter.next();
			if (obj.obj instanceof AttributeDef) {
				AttributeDef attr = (AttributeDef) obj.obj;
				attrv.add(attr);
			}
			if(obj.obj instanceof RoleDef){
				RoleDef role = (RoleDef) obj.obj;
				// not an embedded role and roledef not defined in a lightweight association?
				if (!obj.embedded && !((AssociationDef)v).isLightweight()){
					attrv.add(role);
				}
				// a role of an embedded association?
				if(obj.embedded){
					AssociationDef roleOwner = (AssociationDef) role.getContainer();
					if(roleOwner.getDerivedFrom()==null){
						attrv.add(role);
					}
				}
			}
		}
	}
	private void addCoord(IFMEFeature ret,IomObject value){
		if(value!=null){
			String c1=value.getattrvalue("C1");
			String c2=value.getattrvalue("C2");
			String c3=value.getattrvalue("C3");
			double xCoord = Double.parseDouble(c1);
			double yCoord = Double.parseDouble(c2);
			if(c3==null){
				ret.setDimension(IFMEFeature.FME_TWO_D);
				ret.add2DCoordinate(xCoord, yCoord);
			}else{
				double zCoord = Double.parseDouble(c3);
				ret.setDimension(IFMEFeature.FME_THREE_D);
				ret.add3DCoordinate(xCoord, yCoord, zCoord);
			}
		}
	}
	private void addArc(IFMEFeature ret,IomObject value){
		if(value!=null){
			String c1=value.getattrvalue("C1");
			String c2=value.getattrvalue("C2");
			String c3=value.getattrvalue("C3");
			double xCoord = Double.parseDouble(c1);
			double yCoord = Double.parseDouble(c2);
			if(c3==null){
				ret.setDimension(IFMEFeature.FME_TWO_D);
				ret.add2DCoordinate(xCoord, yCoord);
			}else{
				double zCoord = Double.parseDouble(c3);
				ret.setDimension(IFMEFeature.FME_THREE_D);
				ret.add3DCoordinate(xCoord, yCoord, zCoord);
			}
		}
	}
	private void setPolyline(IFMEFeature ret,IomObject obj,boolean isSurfaceOrArea){
		if(obj==null){
			return;
		}
		// is POLYLINE?
		if(!isSurfaceOrArea){
			ret.setGeometryType(IFMEFeature.FME_GEOM_LINE);
		}
		if(isSurfaceOrArea){
			IomObject lineattr=obj.getattrobj("lineattr",0);
			if(lineattr!=null){
				//writeAttrs(out,lineattr);
				EhiLogger.logAdaption("Lineattributes not supported by FME; ignored");							
			}
		}
		boolean clipped=obj.getobjectconsistency()==iom_javaConstants.IOM_INCOMPLETE;
		for(int sequencei=0;sequencei<obj.getattrvaluecount("sequence");sequencei++){
			if(clipped){
				//out.startElement(tags::get_CLIPPED(),0,0);
			}else{
				// an unclipped polyline should have only one sequence element
				if(sequencei>0){
					EhiLogger.logError("unclipped polyline with multi 'sequence' elements");
					break;
				}
			}
			IomObject sequence=obj.getattrobj("sequence",sequencei);
			for(int segmenti=0;segmenti<sequence.getattrvaluecount("segment");segmenti++){
				IomObject segment=sequence.getattrobj("segment",segmenti);
				if(segment.getobjecttag().equals("COORD")){
					// COORD
					addCoord(ret,segment);
				}else if(segment.getobjecttag().equals("ARC")){
					// ARC
					addArc(ret,segment);
				}else{
					// custum line form
					EhiLogger.logAdaption("custom line form not supported by FME; ignored");
					//out.startElement(segment->getTag(),0,0);
					//writeAttrs(out,segment);
					//out.endElement(/*segment*/);
				}

			}
			if(clipped){
				//out.endElement(/*CLIPPED*/);
			}
		}
	}
	private void setSurface(IFMEFeature ret,IomObject obj){
		if(obj==null){
			return;
		}
		ret.setGeometryType(IFMEFeature.FME_GEOM_DONUT);
		//IFMEFeatureVector bndries=session.createFeatureVector();
		boolean clipped=obj.getobjectconsistency()==iom_javaConstants.IOM_INCOMPLETE;
		for(int surfacei=0;surfacei<obj.getattrvaluecount("surface");surfacei++){
			if(clipped){
				//out.startElement("CLIPPED",0,0);
			}else{
				// an unclipped surface should have only one surface element
				if(surfacei>0){
					EhiLogger.logError("unclipped surface with multi 'surface' elements");
					break;
				}
			}
			IomObject surface=obj.getattrobj("surface",surfacei);
			for(int boundaryi=0;boundaryi<surface.getattrvaluecount("boundary");boundaryi++){
				IomObject boundary=surface.getattrobj("boundary",boundaryi);
				//IFMEFeature fmeLine=session.createFeature();
				for(int polylinei=0;polylinei<boundary.getattrvaluecount("polyline");polylinei++){
					IomObject polyline=boundary.getattrobj("polyline",polylinei);
					setPolyline(ret,polyline,true);
				}
				//bndries.append(fmeLine);
			}
			if(clipped){
				//out.endElement(/*CLIPPED*/);
			}
		}
		try{
			//int bndc=bndries.entries();
			//EhiLogger.debug("number of boundaries "+Integer.toString(bndc));
			//ret.makeDonuts(bndries,true);
			//fmeLog.logFeature(ret,IFMELogFile.FME_INFORM,-1);
			if(surfaceAsPolyline){
				ret.outerShell();
			}
		}catch(FMEException ex){
			EhiLogger.logError(ex);
		}
	}
	/** list of not yet processed struct values
	 */
	private ArrayList structQueue=null;
	private void enqueStructValue(int parentSqlId,String parentSqlType,String parentSqlAttr,IomObject struct,int structi)
	{
		structQueue.add(new StructWrapper(parentSqlId,parentSqlType,parentSqlAttr,struct,structi));
	}	
	private FmeLogListener listener=null;
	//private Log2File listener2=null;
	private void setupLogging(IFMELogFile logFile)
	{
		if(listener==null){
			listener=new FmeLogListener(logFile);
			EhiLogger.getInstance().addListener(listener);
			//listener2=new Log2File("c:/tmp/ili2fme/ce.log");
			//EhiLogger.getInstance().addListener(listener2);
			EhiLogger.getInstance().addListener(ch.ehi.basics.logging.StdListener.getInstance());
			EhiLogger.getInstance().removeListener(ch.ehi.basics.logging.StdListener.getInstance());
			//EhiLogger.getInstance().setTraceFiler(false);
		}
		
	}
	private boolean doSetupIom=true;
	private void setupIom(){
		// setup iom
		if(doSetupIom){
			ch.interlis.iom.Iom.setLibraryHint(Ili2fme.ILI2FME_JAR);
			ch.interlis.iom.Iom.init();
			IomErrorListener iomerrs=new IomErrorListener();
			ch.interlis.iom.Iom.addErrListener(iomerrs);
			doSetupIom=false;
		}
	}
	private void cleanup(){
		//EhiLogger.debug("cleanup");
		// did setup iom?
		if(!doSetupIom){
			//EhiLogger.debug("call Iom.end()");
			ch.interlis.iom.Iom.end();
			doSetupIom=true;
		}
		if(listener!=null){
			EhiLogger.getInstance().removeListener(listener);
			listener=null;
		}
		//if(listener2!=null){
		//	EhiLogger.getInstance().removeListener(listener2);
		//	listener2=null;
		//}
	}
}
