/* This file is part of the ili2fme project.
 * For more information, please see <http://www.interlis.ch>.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
package ch.interlis.ili2fme;

import ch.interlis.iom.ErrorListener;
import ch.interlis.iom.*;
import ch.interlis.iom.swig.iom_javaConstants;
import ch.ehi.basics.logging.EhiLogger;

/** Bridge to log IOM log-events to the EhiLogger system. 
 *  Called by iom.
 * 	One object should be registered by a call to Iom.addErrListener(...);
 * @author ce
 * @version $Revision: 1.0 $ $Date: 02.11.2005 $
 */
public class IomErrorListener implements ErrorListener {
	/** callback called by iom.
	 */
	public void issueError(IomObject errobj)
	{
		StringBuffer msg=new StringBuffer();
		String objtag=errobj.getobjecttag();
		if(objtag.equals("iomerr04.errors.Error")){
			msg.append(errobj.getattrvalue("message"));
		}else if(objtag.equals("iomerr04.errors.XmlParseError")){
			msg.append(errobj.getattrvalue("kind"));msg.append(", ");
			msg.append(errobj.getattrvalue("line"));msg.append(", ");
			msg.append(errobj.getattrvalue("col"));msg.append(": ");
			msg.append(errobj.getattrvalue("message"));
		}else if(objtag.equals("iomerr04.errors.SemanticError")){
			msg.append("basket ");msg.append(errobj.getattrvalue("bid"));
			String oid=errobj.getattrvalue("oid");
			if(oid!=null){
				msg.append(", object ");msg.append(oid);
			}
			msg.append(": ");
			msg.append(errobj.getattrvalue("message"));
		}else{
			msg.append("ERROR: ");msg.append(objtag);
			// dump all primitive type attributes of unknown object
			for(int i=0;i<errobj.getattrcount();i++){
			   String attr=errobj.getattrname(i);
			   String value=errobj.getattrvalue(attr);
			   if(value!=null){
				msg.append("; ");
				msg.append(attr);
				msg.append(" ");
				msg.append(value);
			   }
			}
		}
		EhiLogger.logError(msg.toString());		
	}
}
