/*============================================================================

   Name     : red.java

   System   : FME Developer Kit

   Language : Java

   Purpose  : Implements the IFMEReaderCreator and IFMEWriterCreator
              interfaces for the Red format (near-identical to
              Arc/Info Generate).

   Author               Date            Changes made
   ------------------   ------------    -------------------------------
   Your Name            Xxx XX, 200X    Original definition

                  Copyright (c) 1999 - 2000, Your Company Goes Here
                              All Rights Reserved

    The entire risk as to the results and performance of the software,
    supporting text and other information contained in this file
    (collectively called the "Software") is with the user.  Although
    Safe Software Incorporated has used considerable efforts in preparing
    the Software, Safe Software Incorporated does not warrant the
    accuracy or completeness of the Software. In no event will Safe Software
    Incorporated be liable for damages, including loss of profits or
    consequential damages, arising out of the use of the Software.

    NOTE: Replace the above header with whatever is appropriate for your
    company.

============================================================================*/

package com.mycompany.myformat;

import com.mycompany.myformat.myreader.RedReader;
import com.mycompany.myformat.mywriter.RedWriter;
import COM.safe.fme.pluginbuilder.*;
import COM.safe.fmeobjects.*;
import java.io.*;
import java.util.*;

public final class red implements IFMEReaderCreator, IFMEWriterCreator
{
   // -----------------------------------------------------------------------
   // createReader
   //
   // The createReader method returns a new IFMEReader object.  It is given
   // FME Mapping File object, the FME Logfile object, the FME Coordinate
   // System Manager object and the FMEObjects Session object.
   //---------------------------------------------------------------------

  /**
  Creates a new reader object (that implements the IFMEReader interface).
  Note that any Exception thrown in this method will
  cause a message to be displayed on standardout,
  in addition to the Exception's specific message text. As well, both
  messages will be displayed in the log file.

  @param mappingFile the mapping file object to be used by this IFMEReader.
  @param logFile the log file object to be used by this IFMEReader.
  @throws Exception if a problem is encountered during creation of the reader.
  */
   public IFMEReader createReader(IFMEMappingFile mappingFile,
                                  IFMELogFile logFile,
                                  IFMECoordSysManager coordSysMan,
                                  IFMESession session,
                                  String readerTypeName,
                                  String readerKeyword)
                                  throws Exception
   {
      RedReader readerInstance = new RedReader(readerTypeName, readerKeyword);
      readerInstance.init(mappingFile, logFile, coordSysMan, session);
      return readerInstance;
   }



   // -----------------------------------------------------------------------
   // createWriter
   //
   // The createWriter method returns a new IFMEWriter object.  It is given
   // FME Mapping File object, the FME Logfile object, the FME Coordinate
   // System Manager object and the FMEObjects Session object.
   //---------------------------------------------------------------------

  /**
  Creates a new writer object (that implements the IFMEWriter interface).
  Note that any Exception thrown in this method will
  cause a message to be displayed on standardout,
  in addition to the Exception's specific message text. As well, both
  messages will be displayed in the log file.

  @param mappingFile the mapping file object to be used by this IFMEWriter.
  @param logFile the log file object to be used by this IFMEWriter.
  @throws Exception if a problem is encountered during creation of the reader.
  */
   public IFMEWriter createWriter(IFMEMappingFile mappingFile,
                                  IFMELogFile logFile,
                                  IFMECoordSysManager coordSysMan,
                                  IFMESession session,
                                  String writerTypeName,
                                  String writerKeyword)
                                  throws Exception
   {
      RedWriter writerInstance = new RedWriter(writerTypeName, writerKeyword);
      writerInstance.init(mappingFile, logFile, coordSysMan, session);
      return writerInstance;
   }
}


